import os

from threading import RLock

from pytagsfs.exceptions import (
  PathNotFound,
  InvalidArgument,
)
from pytagsfs.util import (
  nonatomic,
  attr_named,
)


class File(object):
    _lock = None
    filesystem = None
    fake_path = None
    real_path = None
    flags = None
    truncate_to = None

    def __init__(
      self,
      filesystem,
      fake_path,
      flags,
      truncate_to = None,
    ):
        self._lock = RLock()
        self.filesystem = filesystem
        self.fake_path = fake_path
        self.real_path = filesystem.source_tree_rep.get_real_path(fake_path)
        self.flags = flags
        self.truncate_to = truncate_to

    def fgetattr(self):
        raise NotImplementedError

    def flush(self):
        raise NotImplementedError

    def fsync(self, datasync):
        raise NotImplementedError

    def ftruncate(self, len):
        raise NotImplementedError

    def read(self, length, offset):
        raise NotImplementedError

    def release(self, flags):
        pass

    def write(self, buf, offset):
        raise NotImplementedError


class ReadOnlyFile(File):
    fd = None
    file = None

    def __init__(self, *args, **kwargs):
        super(ReadOnlyFile, self).__init__(*args, **kwargs)
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        real_path = self.filesystem.encode_real_path(self.real_path)
        self.file = os.fdopen(os.open(real_path, self.flags), 'r')

################################################################################

    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        stat_result = os.fstat(self.file.fileno())
        st_size = stat_result.st_size
        if (self.truncate_to is not None) and (st_size > self.truncate_to):
            st_size = self.truncate_to
        return os.stat_result((
          stat_result.st_mode,
          stat_result.st_ino,
          stat_result.st_dev,
          stat_result.st_nlink,
          stat_result.st_uid,
          stat_result.st_gid,
          st_size,
          stat_result.st_atime,
          stat_result.st_mtime,
          stat_result.st_ctime,
        ))

    def ftruncate(self, len):
        raise InvalidArgument

    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        if self.truncate_to is not None:
            length = self.truncate_to - offset
            if length < 0:
                length = 0
        self.file.seek(offset)
        return self.file.read(length)

    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        self.file.close()

    def write(self, buf, offset):
        raise InvalidArgument


class ReadWriteFile(File):
    fd = None

    def __init__(self, *args, **kwargs):
        super(ReadWriteFile, self).__init__(*args, **kwargs)
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        real_path = self.filesystem.encode_real_path(self.real_path)
        self.fd = os.open(real_path, self.flags)

        if self.truncate_to is not None:
            os.ftruncate(self.fd, self.truncate_to)

################################################################################

    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        os.lseek(self.fd, offset, 0)
        return os.read(self.fd, length)

    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        return os.close(self.fd)

    @nonatomic(attr_named('_lock'))
    def write(self, buf, offset):
        os.lseek(self.fd, offset, 0)
        return os.write(self.fd, buf)

    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        return os.fstat(self.fd)

    @nonatomic(attr_named('_lock'))
    def flush(self):
        return os.close(os.dup(self.fd))

    @nonatomic(attr_named('_lock'))
    def fsync(self, datasync):
        if datasync and hasattr(os, 'fdatasync'):
            return os.fdatasync(self.fd)
        else:
            return os.fsync(self.fd)

    @nonatomic(attr_named('_lock'))
    def ftruncate(self, len):
        return os.ftruncate(self.fd, len)
